import numpy as np
from ._base import ExtractiveSummarizer


class Rouge1Summarizer(ExtractiveSummarizer):
    """Assign a higher importance score based on ROUGE1 score of the sentences within the document.

    metric : {'f1', 'precision','recall'}, default='f1'
        Metric to be used for ROUGE1 computation.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return L2 normalized score vector.
    """

    def __init__(self, metric='f1', normalize=True):
        self.normalize = normalize
        if metric not in ['f1', 'precision', 'recall']:
            raise Exception(f"mode should be one of 'f1', 'precision','recall'")

        self.metric = metric

    def predict(self, sentences):
        scores = np.array([sent.rouge1(self.metric) for sent in sentences])

        if self.normalize:
            return scores / scores.sum()
        else:
            return scores



from collections import defaultdict
from math import ceil
from typing import List
import click
from tabulate import tabulate
import numpy as np
from sklearn.metrics import ndcg_score
from tqdm import tqdm
from ..dataset import load_annotated_corpus
from ..summarize import RandomSummarizer, PositionSummarizer, Rouge1Summarizer, KMeansSummarizer, AutoKMeansSummarizer, \
    DecomposedKMeansSummarizer
from sadedegel.bblock import Sentences, Doc


def to_sentence_list(sents: List[str]) -> List[Sentences]:
    l = []

    for i, sent in enumerate(sents):
        l.append(Sentences(i, sent, l))

    return l


@click.group()
def cli():
    pass


@cli.command()
@click.option("-f", "--table-format", default="github")
def evaluate(table_format):
    anno = load_annotated_corpus()

    scores = defaultdict(list)
    for name, summarizer in [('Random', RandomSummarizer()), ('FirstK', PositionSummarizer()),
                             ('LastK', PositionSummarizer('last')), ('Rouge1 (f1)', Rouge1Summarizer()),
                             ('Rouge1 (precision)', Rouge1Summarizer('precision')),
                             ('Rouge1 (recall)', Rouge1Summarizer('recall')),
                             ('KMeans', KMeansSummarizer()),
                             ('AutoKMeansSummarizer', AutoKMeansSummarizer()),
                             ('DecomposedKMeansSummarizer', DecomposedKMeansSummarizer())]:
        for doc in anno:
            y_true = [doc['relevance']]

            sents_list = to_sentence_list(doc['sentences'])

            if name in ("KMeans", 'AutoKMeansSummarizer', 'DecomposedKMeansSummarizer'):
                y_pred = [summarizer.predict(Doc(None, doc['sentences']))]
            else:
                y_pred = [summarizer.predict(sents_list)]

            score_10 = ndcg_score(y_true, y_pred, k=ceil(len(doc['sentences']) * 0.1))
            score_50 = ndcg_score(y_true, y_pred, k=ceil(len(doc['sentences']) * 0.5))
            score_80 = ndcg_score(y_true, y_pred, k=ceil(len(doc['sentences']) * 0.8))

            scores[name].append((score_10, score_50, score_80))

    table = [[algo, np.array([s[0] for s in scores]).mean(), np.array([s[1] for s in scores]).mean(),
              np.array([s[2] for s in scores]).mean()] for
             algo, scores in scores.items()]

    # TODO: Sample weigth of instances.
    print(
        tabulate(table, headers=['Method', 'ndcg(k=0.1)', 'ndcg(k=0.5)', 'ndcg(k=0.8)'], tablefmt=table_format))


if __name__ == '__main__':
    cli()



from sklearn.linear_model import PassiveAggressiveClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction import FeatureHasher
from sklearn.feature_extraction import DictVectorizer

from sklearn.pipeline import Pipeline
from os.path import dirname
from pathlib import Path
from joblib import dump, load
from loguru import logger


def create_model():
    """Creates a new sbd model detector."""
    return Pipeline([('feat', DictVectorizer()), ('dt', RandomForestClassifier())])


def save_model(model, name="sbd.pickle"):
    model_file = (Path(dirname(__file__)) / 'model' / name).absolute()

    dump(model, model_file)


def load_model(name="sbd.pickle"):
    model_file = (Path(dirname(__file__)) / 'model' / name).absolute()
    logger.info(f"Loading sbd model from {model_file}")

    return load(model_file)



from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from math import ceil
from ._base import ExtractiveSummarizer
from ..bblock import Doc


class KMeansSummarizer(ExtractiveSummarizer):

    def __init__(self, n_clusters=2, random_state=42, normalize=True):
        self.normalize = normalize
        self.n_clusters = n_clusters
        self.random_state = random_state

    def predict(self, doc):
        if type(doc) != Doc:
            raise Exception(f"KMeansSummarizer.predict accepts only Doc type because of bert_embeddings dependency.")

        effective_n_clusters = min(self.n_clusters, len(doc))

        scores = 1 / (KMeans(n_clusters=effective_n_clusters, random_state=self.random_state).fit_transform(
            doc.bert_embeddings).min(axis=1) + 1e-10)

        if self.normalize:
            return scores / scores.sum()
        else:
            return scores


class AutoKMeansSummarizer(ExtractiveSummarizer):
    """Kmeans cluster automatically deciding on the number of clusters to be used based on document length."""

    def __init__(self, n_cluster_to_length=0.05, min_n_cluster=2, random_state=42, normalize=True):
        self.normalize = normalize
        self.n_cluster_to_length = n_cluster_to_length
        self.min_n_cluster = min_n_cluster
        self.random_state = random_state

    def predict(self, doc):
        if type(doc) != Doc:
            raise Exception(f"KMeansSummarizer.predict accepts only Doc type because of bert_embeddings dependency.")

        effective_n_clusters = min(max(ceil(len(doc) * self.n_cluster_to_length), self.min_n_cluster), len(doc))

        scores = 1 / (KMeans(n_clusters=effective_n_clusters, random_state=self.random_state).fit_transform(
            doc.bert_embeddings).min(axis=1) + 1e-10)

        if self.normalize:
            return scores / scores.sum()
        else:
            return scores


class DecomposedKMeansSummarizer(ExtractiveSummarizer):
    """BERT embeddings are high in dimension and potentially carry redundant information that can cause
        overfitting or curse of dimensionality effecting in clustering embeddings.

        DecomposedKMeansSummarizer adds a PCA step (or any othe lsinear/non-linear dimensionality reduction technique)
         before clustering to obtain highest variance in vector fed into clustering
    """

    def __init__(self, n_clusters=2, n_components=48, random_state=42, normalize=True):
        self.normalize = normalize
        self.n_clusters = n_clusters
        self.n_components = n_components
        self.random_state = random_state

    def predict(self, doc):
        if type(doc) != Doc:
            raise Exception(f"KMeansSummarizer.predict accepts only Doc type because of bert_embeddings dependency.")

        effective_n_clusters = min(self.n_clusters, len(doc))
        effective_n_components = min(self.n_components, len(doc))

        pipeline = Pipeline(
            [('pca', PCA(effective_n_components)),
             ('kmeans', KMeans(effective_n_clusters, random_state=self.random_state))])

        scores = 1 / (pipeline.fit_transform(doc.bert_embeddings).min(axis=1) + 1e-10)

        if self.normalize:
            return scores / scores.sum()
        else:
            return scores



from typing import List
import numpy as np

from ..bblock import Doc, Sentences
from ._base import ExtractiveSummarizer


class CompositeSummarizer(ExtractiveSummarizer):
    def predict(self, sents: List[Sentences]) -> np.ndarray:
        pass



from typing import List, Union
import numpy as np
from abc import ABC, abstractmethod
from ..bblock import Doc, Sentences


class ExtractiveSummarizer(ABC):

    @abstractmethod
    def predict(self, sents: List[Sentences]) -> np.ndarray:
        pass

    def __call__(self, sents: Union[Doc, List[Sentences]], k: int) -> List[Sentences]:
        if type(sents) == Doc:
            sents = sents.sents  # Get the list of Sentences from Doc

        scores = self.predict(sents)
        topk_inds = np.argpartition(scores, k)[-k:]  # returns indices of k top sentences
        topk_inds.sort()  # fix the order of sentences

        summ = [sents[i] for i in topk_inds]
        return summ



from fastapi.openapi.utils import get_openapi

import uvicorn
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from pydantic.typing import List

from sadedegel.bblock import Doc, Sentences
from sadedegel.summarize import RandomSummarizer, PositionSummarizer, Rouge1Summarizer
import click
import numpy as np
from sadedegel.about import __version__
from enum import Enum
from loguru import logger


class TimeUnitEnum(str, Enum):
    SECOND = 'second'
    MINUTE = 'minute'


class BasicRequest(BaseModel):
    doc: str
    wpm: int = 170


class Request(BasicRequest):
    duration: int = 5
    unit: TimeUnitEnum = TimeUnitEnum.MINUTE


class DocSummary(BaseModel):
    sentence_count: int
    word_count: int


class RichSummary(DocSummary):
    wpm: int
    duration: float
    unit: TimeUnitEnum


class Response(BaseModel):
    sentences: List[str]
    original: DocSummary
    summary: DocSummary


class APIInfo(BaseModel):
    version: str
    email: str
    website: str
    github: str


app = FastAPI()


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="sadedeGel",
        version=__version__,
        description="sadedeGel is an extraction-based summarizer for Turkish news content",
        routes=app.routes,
    )
    openapi_schema["info"]["x-logo"] = {
        "url": "https://sadedegel.ai/dist/img/logo-2.png?s=280&v=4"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

app.add_middleware(
    CORSMiddleware,
    allow_origins='*',
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def summary_filter(sents, scores, word_count, limit=None):
    rank = np.argsort(scores)[::-1]
    logger.info(rank)

    logger.info(limit)

    if limit:
        return sents[rank[word_count[rank].cumsum() <= limit]]
    else:
        return sents[rank]


@app.get('/')
async def home():
    return RedirectResponse("http://sadedegel.ai")


@app.get("/api/info", tags=["Information"], summary="sadedeGel metadata")
async def info():
    return APIInfo(version=__version__, email="info@sadedegel.ai", website="sadedegel.ai",
                   github="https://github.com/GlobalMaksimum/sadedegel")


def summarize(summarizer, sentences: List[Sentences], limit: float) -> Response:
    word_count = np.array([len(s) for s in sentences], dtype=np.int)

    scores = summarizer.predict(sentences)

    logger.info(scores)
    logger.info(word_count)

    sentences_limited = summary_filter(np.array(sentences), scores, word_count, limit)

    logger.info(sentences_limited.tolist())

    return Response(sentences=[s.text for s in sentences_limited.tolist()],
                    original=DocSummary(sentence_count=len(sentences), word_count=word_count.sum()),
                    summary=DocSummary(sentence_count=len(sentences_limited),
                                       word_count=np.array([len(s) for s in sentences_limited.tolist()],
                                                           dtype=np.int).sum()))


@app.post("/api/doc/statistics", tags=["Utility"], summary="Calculate approximate reading duration of a document")
async def duration(req: BasicRequest):
    """Calculates the approximate reading duration for the document based on reader read speed"""

    sentences = Doc(req.doc).sents

    word_count = sum((len(s) for s in sentences))

    dur = (word_count / (req.wpm / 60))
    unit = TimeUnitEnum.SECOND

    logger.info(f"Total duration {dur}")

    return RichSummary(sentence_count=len(sentences), word_count=word_count, wpm=req.wpm, duration=dur, unit=unit)


@app.post("/api/summarizer/random", tags=["Summarizer"], summary="Use RandomSummarizer")
async def random(req: Request):
    """Baseline [Random Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Picks up random sentences until total number of tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(RandomSummarizer(seed=None), sentences, req.wpm * duration_in_min)


@app.post("/api/summarizer/firstk", tags=["Summarizer"],
          summary="Use Position Summarizer to obtain a FirstK summarizer")
async def firstk(req: Request):
    """Baseline [Position Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Picks up first a few sentences until total number of tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(PositionSummarizer(), sentences, req.wpm * duration_in_min)


@app.post("/api/summarizer/rouge1", tags=["Summarizer"],
          summary="Use unsupervised Rouge1 Summarizer")
async def firstk(req: Request):
    """Baseline [Rouge1 Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Rank sentences based on their rouge1 score in Document and return a list of sentences until number of total tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(Rouge1Summarizer(), sentences, req.wpm * duration_in_min)


@click.command()
@click.option("--host", '-h', help="Hostname", default="0.0.0.0")  # nosec
@click.option("--log-level",
              type=click.Choice(['debug', 'info'], case_sensitive=False), help="Logging Level", default="info")
@click.option("--reload", is_flag=True, default=False, help="enable/disable auto reload for development.")
@click.option("--port", help="Port", default=8000)
def server(host, log_level, reload, port):
    """Span a sadedeGel http server instance."""

    if reload:
        uvicorn.run("sadedegel.server.__main__:app", host=host, log_level=log_level, reload=reload, port=port)
    else:
        uvicorn.run(app, host=host, log_level=log_level, reload=reload, port=port)


if __name__ == "__main__":
    server()
